#!/usr/bin/env python

# -*- coding: utf-8 -*-

"""ospi-client: performs operation for opsi clients on opsi server via JSON-RPC."""

__author__ = "Joerg Steffens"
__copyright__ = "Copyright 2012-2015, dass IT GmbH"
__license__ = "GPL"
__version__ = "1.1"
__email__ = "joerg.steffens@dass-it.de"

#self.command("opsi-admin -d method host_createOpsiClient "+ \
        #computername + " null " + "\\'"+description+"\\'" + \
        #" \\'created by dassadmin\\' " + mac_address + " " + \
        #ip_address)
#self.command("opsi-admin -d method configState_create clientconfig.depot.id " + \
        #computername + " " + depotName)

import argparse
import jsonrpc
import logging
import os
from pprint import pprint, pformat
import time

UrlJsonRpc="https://<username>:<password>@opsi:4447/rpc"

HelpEpilog="WARNING: python-json-rpc is known to have problems with HTTP proxies. In case of problems, make sure, the environment variables http_proxy and/or https_proxy are *not* set."

class OpsiRpc:

    UrlJsonRpcDefault="https://opsi:4447/rpc"

    ProductAttributesCopy = ['actionRequest','actionResult','installationStatus','packageVersion','productVersion']

    def __init__(self, urlJsonRpc = UrlJsonRpcDefault, debug=False ):
        self.logger=logging.getLogger(__name__)
        self.debug=debug
        self.urlJsonRpc=urlJsonRpc
        self.rpc=jsonrpc.ServiceProxy(self.urlJsonRpc)
        self.logger.debug( "initialized: " + self.urlJsonRpc )

    def list(self):
        print( "\n".join( self.rpc.getClientIds_list() ) )
        return True


    def getClientsWithProduct( self, product ):
        return self.rpc.productOnClient_getObjects( [], { "productId": product, "installationStatus": "installed" } )


    def listClients( self, product ):
        if product:
            for client in self.getClientsWithProduct( product ):
                print client['clientId']
        else:
            return self.list()
        return True

    def exists(self, src):
        return len( self.rpc.host_getObjects( [], {"id":src} ) ) == 1

    def info(self, src):
        if not self.exists( src ):
            print "failed: opsi client", src, "does not exist"
            return False
        print src + ":"
        host = self.rpc.host_getHashes( [], {"id":src} )[0]
        print "  IP:", host["ipAddress"]
        print "  MAC:", host["hardwareAddress"]
        print "  inventory:", host["inventoryNumber"]
        print "  last seen:", host["lastSeen"]
        print "  notes:", host["notes"]
        print "  depot:", self.clientGetDepot( src )

        print host
        print "-->"
        print self.rpc.configState_getHashes([], { "id":src})
        print "<--"

        print "  products:"
        products = self.getProductOnClient( src, [] )
        for i in products:
            print "    " + i['productId'] + ":"
            print "      " + i['installationStatus'], "(",
            if i['actionRequest']:
                print i['actionRequest'],
                if i['actionProgress']:
                    print i['actionProgress'],
            print ")"
            print "      ",
            pprint( i, indent=8 )
        print "  propertyState:"
        for p in  self.getProductPropertyState(src):
            print "    %-16s %24s: %s"  % ( p['productId'],p['propertyId'], p['values'] )
        return True

    def clean(self, src):
        if not self.exists( src ):
            return False
        products = self.rpc.productOnClient_getObjects( [], { 'clientId': src } )
        self.rpc.productOnClient_deleteObjects( products )

        products = self.rpc.productPropertyState_getObjects( [], { 'objectId': src } )
        self.rpc.productPropertyState_deleteObjects( products )

        if self.debug:
            pprint( self.getProductOnClient( src ) )
        return True

    def getOpsiConfigserverId(self):
        # there should always be only one OpsiConfigserver
        opsiConfigservers=self.rpc.host_getHashes( [], { "type": "OpsiConfigserver" } )
        try:
            return opsiConfigservers[0]['id']
        except (KeyError,IndexError) as e:
            self.logger.error( "failed to retreive OpsiConfigserver" )

    def createClient(self, name, opsiHostKey, description, notes, hardwareAddress, ipAddress):
        #    self.rpc.host_createOpsiClient( name, opsiHostKey, description, notes, hardwareAddress, ipAddress )
        self.updateClient( name, opsiHostKey, description, notes, None, hardwareAddress, ipAddress )

    def deleteClient(self, name):
        self.rpc.host_delete( name )

    def renameClient(self, src, dest):
        self.rpc.host_renameOpsiClient(src, dest)

    def updateClient(self, src, opsiHostKey = None, description = None, notes = None, inventoryNumber = None, hardwareAddress = None, ipAddress = None, depot = None ):
        obj = {
          "id" : src,
          "type" : "OpsiClient",
        }
        if opsiHostKey:
            obj['opsiHostKey'] = opsiHostKey
        if description:
            obj['description'] = description
        if notes:
            obj['notes'] = notes
        if inventoryNumber:
            obj['inventoryNumber'] = inventoryNumber
        if hardwareAddress:
            obj['hardwareAddress'] = hardwareAddress
        if ipAddress:
            obj['ipAddress'] = ipAddress

        if self.exists( src ):
            self.rpc.host_updateObject(obj)
        else:
            self.rpc.host_insertObject(obj)

        if depot:
            self.clientSetDepot(src,depot)
        return True

    def clientGetDepot(self, name):
        depot = self.rpc.configState_getHashes( [], {
            "configId": "clientconfig.depot.id",
            "objectId": name } )
        try:
            return depot[0]["values"][0]
        except (IndexError,KeyError):
            return self.getOpsiConfigserverId()

    def clientSetDepot(self, name, depot):
        self.rpc.configState_create( "clientconfig.depot.id", name, depot )

    def copyClient( self, src, dst, ipAddress = None, hardwareAddress = None, depot = None, description = "", copyProperties = True ):

        print "create/update", dst, "from template", src + ":",
        obj = {
          "id" : dst,
          "type" : "OpsiClient",
          "notes" : "copy of " + src,
          "description" : description,
          #"inventoryNumber" : "",
        }
        if hardwareAddress:
            obj['hardwareAddress'] = hardwareAddress
        if ipAddress:
            obj['ipAddress'] = ipAddress

        if self.exists( dst ):
            self.rpc.host_updateObject(obj)
        else:
            self.rpc.host_insertObject(obj)

        if depot:
            self.clientSetDepot(dst,depot)

        self.copyConfigState(src, dst)
#        self.copyConfigServer(src, dst)

        if self.debug:
            pprint( self.getProductOnClient( src ) )
        self.copyProductOnClient( src, dst )
        if copyProperties:
            if self.debug:
                print "copy product properties"
            if not depot:
                # get default Properties from Master Depot Server (OpsiConfigserver)
                depot = self.getOpsiConfigserverId()
            self.copyProductPropertyState( src, dst, depot )


        if self.debug:
            print "done"

        return True


    def copyConfigState(self, src, dst):
        configIds = [ "clientconfig.dhcpd.filename", "clientconfig.configserver.url", "software-on-demand.product-group-ids" ]

        for confId in configIds:
                result = self.rpc.configState_getObjects([], {
                        "configId": confId,
                        "objectId": src
                })

                if len(result)>0:
                        result = result[0]
                else:
                        continue

                values = result['values']
                
                print "  set configState %s to %s" % (confId, str(values))

                self.rpc.configState_create(confId, dst, values)


#    def copyConfigServer(self, src, dst):
#        configIds = [ "clientconfig.configserver.url" ]
#
#        for confId in configIds:
#                result = self.rpc.configState_getObjects([], {
#                        "configId": confId,
#                        "objectId": src
#                })
#
#                if len(result)>0:
#                        result = result[0]
#                else:
#                        continue
#
#                value = result['values'][0]
#                
#                print "  set configState %s to %s" % (confId, value)
#
#                self.rpc.configState_create(confId, dst, value)

    def getProductOnClient( self, client, attributes = ProductAttributesCopy ):
        return self.rpc.productOnClient_getHashes( [], { 'clientId': client } )

    def copyProductOnClient( self, src, dst, attributes = ProductAttributesCopy ):
        products_src = self.rpc.productOnClient_getHashes( attributes, { 'clientId': src } )
        products_dst = []
        for i in products_src:
            if self.debug: 
                print i['productId']
                pprint( i )
            i['clientId'] = dst
            products_dst.append(i)
        self.rpc.productOnClient_createObjects( products_dst )
        if self.debug:
            pprint( self.getProductOnClient( dst ) )


    def getProductPropertyState( self, client, attributes = [] ):
        return self.rpc.productPropertyState_getHashes( [], { 'objectId': client } )


    def copyProductPropertyState( self, src, dst, default = None, attributes = [] ):
        if default:
            productProperties_default = self.getProductPropertyState( default, attributes )
        else:
            productProperties_default = []

        productProperties_src = self.getProductPropertyState( src, attributes )
        productProperties_dst = []
        for i in productProperties_src:
            use_default=False
            default_value=None
            for j in productProperties_default:
                if i['productId'] == j['productId'] and i["propertyId"] == j["propertyId"]:
                    default_value = j['values']
                    if i['values'] == j['values']:
                        use_default=True
            if self.debug:
                print i['productId'], "-", i["propertyId"] + ": ", pformat(i["values"]),
                if use_default:
                    print "(use default)"
                else:
                    print "(set, default:", default_value, ")"
            if not use_default:
                i['objectId'] = dst
                productProperties_dst.append(i)
        self.rpc.productPropertyState_createObjects( productProperties_dst )
        if self.debug:
            pprint( self.getProductPropertyState( dst ) )


    def getClientProductProperty( self, client, product ):
        return self.rpc.getProductProperties_hash( product, [ client ] )

    def setProductPropertiesOnClient( self, dst, product, properties ):
        self.rpc.setProductProperties(product,properties,dst)

    def setProductPropertyOnClient( self, dst, product, prop, value):
        self.rpc.setProductProperty(product,prop,value,dst)


if __name__ == '__main__':
    logging.basicConfig(format='%(message)s')
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    
    parser = argparse.ArgumentParser(description='Command line tool for OPSI configuration.', epilog=HelpEpilog )

    parser.add_argument( '--debug', action='store_true', help="enable debugging output" )

    parser_url = parser.add_mutually_exclusive_group(required=True)
    parser_url.add_argument( '--url', help="OPSI Server JSON-RPC url, in following format: " + UrlJsonRpc )
    
    parser_url.add_argument( '--server', help="OPSI Server (instead of URL)" )
    username_default=os.getlogin()
    
    parser.add_argument( '--username', help="username (instead of URL), default: " + username_default, default=username_default )
    parser.add_argument( '--password', help="password (instead of URL)" )
    
    subparsers = parser.add_subparsers(title='subcommands',
        description='valid subcommands',
        help='additional help',
        dest='subcommand' )

    parser_clean = subparsers.add_parser('clean', help='remove all product states from a opsi client' )
    parser_clean.add_argument( 'src', help="source opsi client to clean" )

    parser_copy = subparsers.add_parser('copy', help='copy/create a opsi client from a template opsi client')
    parser_copy.add_argument( 'src', help="source/template opsi client" )
    parser_copy.add_argument( 'dst', help="opsi client to be created" )
    parser_copy.add_argument( '--ip', help="IP address of the new opsi client" )
    parser_copy.add_argument( '--mac', help="MAC address of the new opsi client" )
    parser_copy.add_argument( '--depot', help="depot server the new opsi client should be located" )
    #parser_copy.add_argument( '--no-properties', action='store_false', help="don't copy product properties" )

    parser_createBaculaConfigFiles = subparsers.add_parser('createBaculaConfigFiles', help='create Bacula config files for all clients that have bacula installed')

    parser_exists = subparsers.add_parser('exists', help='check, if a opsi clients exists' )
    parser_exists.add_argument( 'src', help="source opsi client" )
    #parser_list = subparsers.add_parser('list', help='list all opsi clients' )

    parser_listClients = subparsers.add_parser('listClients', help='list opsi clients')
    parser_listClients.add_argument( '--product', help="only list clients, that have product installed" )

    parser_info = subparsers.add_parser('info', help='print information about a opsi client' )
    parser_info.add_argument( 'src', help="opsi client" )

    parser_update = subparsers.add_parser('update', help='update/create a opsi client')
    parser_update.add_argument( 'src', help="opsi client to be created/updated" )
    parser_update.add_argument( '--ip', help="IP address of the opsi client" )
    parser_update.add_argument( '--mac', help="MAC address of the opsi client" )
    parser_update.add_argument( '--description', help="a description of the client" )
    parser_update.add_argument( '--notes', help="notes about the client" )
    parser_update.add_argument( '--inventory', help="inventory number" )
    parser_update.add_argument( '--depot', help="depot server the opsi client should be located" )

    args = parser.parse_args()

    if args.debug:
        logger.setLevel(logging.DEBUG)

    url=args.url
    if (not url):
        if args.server:
            account=""
            if args.username and args.password:
                account=args.username + ":" + args.password + "@"
            elif args.username:
                account=args.username + "@"
            url="https://" + account + args.server + ":4447/rpc"
        else:
            parser.error( "argument --url is required" )

    opsi=OpsiRpc( url, args.debug )

    result = True

    try:
        if args.subcommand == "clean":
            result = opsi.clean( args.src )
        elif args.subcommand == "copy": 
            result = opsi.copyClient( args.src, args.dst, args.ip, args.mac, args.depot )
        elif args.subcommand == "exists":
            result = opsi.exists( args.src )
        elif args.subcommand == "list":
            result = opsi.list()
        elif args.subcommand == "listClients":
            result = opsi.listClients( args.product )
        elif args.subcommand == "info":
            result = opsi.info( args.src )
        elif args.subcommand == "update": 
            result = opsi.updateClient( args.src, None, args.description, args.notes, args.inventory, args.mac, args.ip, args.depot )
        else:
            print "not yet implemented"
    except IOError as e:
        result = False
        # connection refused
        print "failed:", e

    if args.debug: print result

    if result:
        exit(0)
    else:
        exit(1)
